# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/cli/cleanup.py: cleanup command entry point

    Copyright (C) 2019-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import time
import sys

import click

from ..core.config import Config
from ..core.i18n import t
from ..core.io import humanize_size
from ..core.lib import is_host_up
from ..core.zpool import Zpool

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@click.command(short_help="cleanup older snapshots")
@click.argument("configfile", type=click.File("r", encoding="utf-8"))
@click.argument("side", default="source", type=str)
def cleanup(configfile, side):

    config = Config.from_fd(configfile)

    assert side in ("source", "target")

    cleanup_side = side
    control_side = "target" if cleanup_side == "source" else "source"

    if cleanup_side == "target":
        click.confirm(
            t(
                "DANGER ZONE: You are about to clean the TARGET. Do you want to continue?"
            ),
            abort=True,
        )
        if config["keep_backlog"].value == True:
            print(t("nothing to do"))
            return

    for side in (cleanup_side, control_side):
        if not is_host_up(side, config):
            print(f'{t("host is not up"):s}: {side:s}')
            sys.exit(1)

    cleanup_zpool = Zpool.from_config(cleanup_side, config=config)
    control_zpool = Zpool.from_config(control_side, config=config)
    available_before = Zpool.available(cleanup_side, config=config)

    transactions = cleanup_zpool.get_cleanup_transactions(control_zpool)

    if len(transactions) == 0:
        print(t("nothing to do"))
        return
    transactions.print_table()

    click.confirm(t("Do you want to continue?"), abort=True)

    transactions.run()

    WAIT = 10
    print(f"waiting {WAIT:d} seconds ...")
    time.sleep(WAIT)
    available_after = Zpool.available(cleanup_side, config=config)
    print(
        (
            f"{humanize_size(available_after, add_color = True):s} available, "
            f"{humanize_size(available_after - available_before, add_color = True):s} freed"
        )
    )



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/configfield.py: Handles configuration fields

    Copyright (C) 2019-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Callable, List, Union

from typeguard import typechecked

from .abc import ConfigFieldABC

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TYPING
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ConfigValueTypes = Union[List[str], str, int, float, bool, None]

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class ConfigField(ConfigFieldABC):
    """
    Mutable.
    """

    def __init__(
        self,
        name: str,
        validate: Callable,
        default: ConfigValueTypes = None,
    ):

        self._name = name
        self._default = default
        self._validate = validate

        if self._default is not None:
            if not self._validate(self._default):
                raise ValueError(f"invalid default value for {self._name}")

        self._value = None

    def __repr__(self) -> str:

        return (
            "<ConfigField "
            f'required="{str(self.required):s}" '
            f'value="{str(self._value):s}" '
            f'default="{str(self._default):s}"'
            ">"
        )

    def copy(self) -> ConfigFieldABC:

        return type(self)(
            name=self._name,
            default=self._default,
            validate=self._validate,
        )

    @property
    def name(self) -> str:

        return self._name

    @property
    def value(self) -> ConfigValueTypes:

        if self._value is not None:
            return self._value

        if self._default is None:
            raise ValueError(f"required value for {self._name} missing")

        return self._default

    @value.setter
    def value(self, value: ConfigValueTypes):

        if self._value is not None:
            raise ValueError(f"value for {self._name} has already been set")
        if value is None:
            return
        if not self._validate(value):
            raise ValueError(f"invalid value for {self._name}")

        self._value = value

    @property
    def valid(self) -> bool:

        return (self._value is not None and self._validate(self._value)) or (
            self._default is not None and self._validate(self._default)
        )

    @property
    def required(self) -> bool:

        return self._default is None

    @property
    def set(self) -> bool:

        return self._value is not None



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/config.py: Handles configuration data

    Copyright (C) 2019-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Dict, TextIO, Union

from typeguard import typechecked
import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import FullLoader as Loader

from .abc import ConfigABC, ConfigFieldABC
from .configspec import CONFIGSPEC

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Config(ConfigABC):
    """
    Immutable.
    """

    def __init__(self, root: Union[str, None] = None, **kwargs: ConfigFieldABC):

        self._root = root
        self._fields = kwargs

    def __repr__(self):

        return "<Config>" if self._root is None else f'<Config root="{self._root:s}">'

    def __getitem__(self, key: str) -> ConfigFieldABC:

        return (
            self._fields[key]
            if self._root is None
            else self._fields[f"{self._root:s}/{key:s}"]
        )

    def group(self, root: str) -> ConfigABC:

        return type(self)(root=root, **self._fields)

    @classmethod
    def _flatten_dict_tree(cls, data: Dict, root: Union[str, None] = None) -> Dict:

        flat_data = {}

        for key, value in data.items():
            if not isinstance(key, str):
                raise TypeError("configuration key is no string", key)
            if root is not None:
                if len(root) > 0:
                    key = f"{root:s}/{key:s}"
            if isinstance(value, dict):
                flat_data.update(cls._flatten_dict_tree(data=value, root=key))
            else:
                flat_data[key] = value

        return flat_data

    @classmethod
    def from_fd(cls, fd: TextIO) -> ConfigABC:

        return cls.from_text(fd.read())

    @classmethod
    def from_text(cls, text: str) -> ConfigABC:

        config = yaml.load(text, Loader=Loader)

        if not isinstance(config, dict):
            raise TypeError("config is no dict", config)

        config_fields = {field.name: field.copy() for field in CONFIGSPEC}
        for key, value in cls._flatten_dict_tree(data=config).items():
            if key not in config_fields.keys():
                raise ValueError("unknown configuration key", key)
            config_fields[key].value = value

        if any((not field.valid for field in config_fields.values())):
            raise ValueError("configuration is not valid")

        return cls(**config_fields)
